#include "../Framework2/Direct3D/Direct3DApp.h"
#include "../Framework2/Util/Model.h"

D3DVERTEXELEMENT9 modelVF[] = {
	{ 0, 0,            D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	{ 0, sizeof(vec3), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,   0 },
	D3DDECL_END()
};

class MainApp : public Direct3DApp {
public:
	void resetCamera();

	bool init();
	bool exit();
	bool load();
	bool unload();

	void drawRoom();
	void drawQuad();
	void drawOutline();
	void drawBlur();
	bool drawFrame();
protected:
	ShaderID depthNormals, outline, blur;

	TextureID colorRT, colorRT2, depthRT;

	Model *model;
	LPDIRECT3DVERTEXDECLARATION9 modelVDecl;
};

void MainApp::resetCamera(){
	position = vec3(700, 0, -700);
	wx = 0;
	wy = PI / 4;
	wz = 0;
}

bool MainApp::init(){
	model = new Model();
	return model->loadFromFile("../Models/Beethoven.hmdl");
}

bool MainApp::exit(){
	delete model;
	return true;
}

bool MainApp::load(){
	if (caps.PixelShaderVersion < D3DPS_VERSION(2,0) || caps.VertexShaderVersion < D3DVS_VERSION(1,1)){
		String error("This demo requires pixel shader 2.0 and vertex shader 1.1.\n"
					 "Your card/driver only supports pixel shader ");
		error.sprintf("%d.%d", (caps.PixelShaderVersion  >> 8) & 255, caps.PixelShaderVersion  & 255);
		error += " and vertex shader ";
		error.sprintf("%d.%d", (caps.VertexShaderVersion >> 8) & 255, caps.VertexShaderVersion & 255);

		addToLog(error);		
		return false;
	}

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	colorRT  = renderer->addRenderTexture(1024, 1024, FORMAT_RGBA8, false);
	colorRT2 = renderer->addRenderTexture(1024, 1024, FORMAT_RGBA8, false);
	depthRT  = renderer->addRenderDepth(1024, 1024, FORMAT_D16);

	if ((depthNormals = renderer->addShader("depthNormals.shd")) == SHADER_NONE) return false;
	if ((outline = renderer->addShader("outline.shd")) == SHADER_NONE) return false;
	if ((blur = renderer->addShader("blur.shd")) == SHADER_NONE) return false;

	dev->CreateVertexDeclaration(modelVF, &modelVDecl);
	return true;
}

bool MainApp::unload(){
	modelVDecl->Release();
	return true;
}

void MainApp::drawRoom(){
	renderer->setShader(depthNormals);
	renderer->apply();

	renderer->changeShaderConstant4x4f("view_proj_matrix", projection * modelView);
	renderer->changeShaderConstant3f("camPos", position);

	dev->SetVertexDeclaration(modelVDecl);
	dev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	Batch *batch = model->getBatch(0);
	dev->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, batch->getVertexCount(), batch->getIndexCount() / 3, batch->getIndices(), D3DFMT_INDEX16, batch->getVertices(), batch->getVertexSize());
}

void MainApp::drawQuad(){
	static float vertices[] = {
		-1,  1, 0, 0, 0,
		 1,  1, 0, 1, 0,
		 1, -1, 0, 1, 1,
		-1, -1, 0, 0, 1,
	};

	dev->SetFVF(D3DFVF_XYZ | D3DFVF_TEXCOORDSIZE2(0) | (1 << D3DFVF_TEXCOUNT_SHIFT));
	dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, 5 * sizeof(float));
}

void MainApp::drawOutline(){
	renderer->setDepthFunc(DEPTH_NONE);
	renderer->setMask(COLOR);
	renderer->setTextures(colorRT);
	renderer->setShader(outline);
	renderer->apply();

	renderer->changeShaderConstant3f("off", 0.13f / length(position));

	drawQuad();
}

void MainApp::drawBlur(){
	renderer->setDepthFunc(DEPTH_NONE);
	renderer->setMask(COLOR);
	renderer->setTextures(colorRT2);
	renderer->setShader(blur);
	renderer->apply();

	drawQuad();
}

bool MainApp::drawFrame(){
	renderer->changeRenderTarget(colorRT, depthRT);
    dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0, 0, 0, 0), 1.0f, 0);

	projection = projectionMatrixX(1.5f, float(height) / float(width), 1, 3000);
	dev->SetTransform(D3DTS_PROJECTION, (D3DMATRIX *) (const float *) transpose(projection));

	modelView = rotateZXY(-wx, -wy, -wz);
	modelView.translate(-position);
	dev->SetTransform(D3DTS_VIEW, (D3DMATRIX *) (const float *) transpose(modelView));

	drawRoom();




	renderer->changeRenderTarget(colorRT2, depthRT);
    dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0, 0, 0, 0), 1.0f, 0);

	drawOutline();




	renderer->changeToDefaultRenderTarget();
    dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0, 0, 0, 0), 1.0f, 0);

	drawBlur();

	return true;
}

App *app = new MainApp();
